# FILE: mmSolve2D.tcl
#
# This file must be interpreted by the mmsolve shell.

Oc_IgnoreTermLoss  ;# Try to keep going, even if controlling terminal
## goes down.

# Support libraries
package require Mms 1.1
package require Net

Oc_Main SetAppName mmSolve2D
Oc_Main SetVersion 1.1.0.1
regexp \\\044Date:(.*)\\\044 {$Date: 1999/08/10 23:57:00 $} _ date
Oc_Main SetDate [string trim $date]
Oc_Main SetAuthor [Oc_Person Lookup dgp]
Oc_Main SetHelpURL [Oc_Url FromFilename [file join [file dirname \
        [file dirname [file dirname [Oc_DirectPathname [info \
        script]]]]] doc userguide 2D_Micromagnetic_Solver_mmS.html]]

Oc_AppendUsageInfo "mmSolve2D [Oc_Main GetVersion]" \
        "OOMMF Two-dimensional Micromagnetic Solver" {
        {-h* {} {Display this help message and quit}}
        {-restart int \
		{1 => use <basename>.log file to restart simulation}}
        {-version {} {Display version information and quit}} }

set restart_option 0
while {[string match -* [lindex $argv 0]]} {
    regexp -- -(.*) [lindex $argv 0] _ option
    set argv [lrange $argv 1 end]
    switch -exact -- $option {
	restart {
	    set param [lindex $argv 0]
	    if {[catch {expr $param && $param} param]} {
                Oc_Log Log "\nERROR: Bad -restart parameter\n" info
		Oc_Log Log "[Oc_GetUsageInfo]" info
                exit
	    }
	    set restart_option $param
	    set argv [lrange $argv 1 end]
	}
        version {
            Oc_Log Log "[Oc_Main GetAppName] [Oc_Main GetVersion]" info
            exit
        }
        default {
           if {[string match h* $option]} {
                Oc_Log Log [Oc_GetUsageInfo] info
                exit
            }
            Oc_Log Log "Unrecognized option: -$option\n[Oc_GetUsageInfo]" panic
            exit 1
        }
    }
}

if {[Oc_Main HasTk]} {
    wm withdraw .        ;# "." is just an empty window.
    package require Ow
    Ow_SetIcon .
} else {
    # Set up to write log messages to oommf/mmsolve2d.errors.
    # Note that the solver class will later re-direct this to its own log
    # file.
    FileLogger SetFile \
	    [file join [Oc_Main GetOOMMFRootDir] mmsolve2d.errors]
    Oc_Log SetLogHandler [list FileLogger Log] panic Oc_Log
    Oc_Log SetLogHandler [list FileLogger Log] error Oc_Log
    Oc_Log SetLogHandler [list FileLogger Log] warning Oc_Log
    Oc_Log SetLogHandler [list FileLogger Log] info Oc_Log
    Oc_Log SetLogHandler [list FileLogger Log] panic
    Oc_Log SetLogHandler [list FileLogger Log] error
    Oc_Log SetLogHandler [list FileLogger Log] warning
    Oc_Log SetLogHandler [list FileLogger Log] info
}

set commands {
    Solver	{{LoadProblem RemoteRelayInput ProblemDescription}}
}

array set inputs {
    ProblemDescription	{ProbEd
			 GetProbDescFile
			 LoadProbDescFile
                        }
}
array set events {}
array set outputs {}

Oc_EventHandler Bindtags LoadProbDescFile LoadProbDescFile
proc LoadProbDescFile { mif_file } {
    global commands events inputs outputs
#    catch { unset inputs } ;###
    catch { unset outputs }  ;# "outputs" is _not_ always set!
    catch { unset events }
    catch { unset commands } 
#    array set inputs {}   ;###
# The commands which can be invoked remotely
set commands {
    Solver	{
		 {Reset ResetThread} 
                 {LoadProblem RemoteRelayInput ProblemDescription}
                 {Run ChangeSolverState Run}  
                 {Relax ChangeSolverState Relax}  
                 {Pause ChangeSolverState Pause}  
		 {Field- FieldAdjust -1}
		 {Field+ FieldAdjust 1}
		}
}

# The events generated by this thread.
array set events {
    Iteration	 {1 solver GetODEStepCount}
    ControlPoint {1 solver GetFieldStepCount}
}

# The quantities which this thread produces as output
#	The last entry should be moved to be set by controller
array set outputs {
    Magnetization	{vectorField
			 {Iteration ControlPoint}
			 solver
			 WriteMagFile
			 datafile
                        }
    TotalField 		{vectorField
			 {Iteration ControlPoint}
			 solver
			 WriteFieldFile
			 datafile
                        }
    DataTable		{DataTable
			 {Iteration ControlPoint}
			 ""
			 MakeDataTable
			 DataTable
			}
}


    global solver protocol restart_option
    mms_mif New mif
    $mif Read $mif_file
    if {![info exists solver]} {
	# First pass---new solver
	mms_solver New solver $mif $restart_option
	Oc_EventHandler Generate LoadProbDescFile InterfaceChange
	# Examine this:
	Oc_EventHandler New _ $solver Delete [list $protocol Delete] \
                -groups [list $solver $protocol]
	Oc_EventHandler New _ $solver RunDone [list RunEndNotify]
        Oc_EventHandler New _ Oc_Main Exit [list $solver Delete] \
                -groups [list $solver]
    } else {
	# Reset existing solver
	ResetThread $mif
    }
    $mif Delete  ;# $solver makes a private copy of $mif
    file delete $mif_file
}

Net_Protocol New protocol -name "OOMMF solver protocol 0.0"
$protocol AddMessage start HasGui {} {
    global solver
    if {![info exists solver]} {
	Oc_EventHandler New _ LoadProbDescFile InterfaceChange \
		[list $connection Tell InterfaceChange] -oneshot 1 \
		-groups [list $connection]
    }
    Oc_EventHandler New _ mms_solver ChangeState \
	    [list $connection Tell Status %state] \
	    -groups [list $connection]
    return [list start [list 0 1]]
}
$protocol AddMessage start Commands {} {
    global commands
    return [list start [list 0 $commands]]
}
$protocol AddMessage start Outputs {} {
    global outputs
    return [list start [list 0 [array get outputs]]]
}
$protocol AddMessage start Inputs {} {
    global inputs
    return [list start [list 0 [array get inputs]]]
}
$protocol AddMessage start Events {} {
    global events
    return [list start [list 0 [array get events]]]
}
$protocol AddMessage start SetHandler {args} {
    return [list start [list 0 [eval SetHandler $args]]]
}
$protocol AddMessage start SetInput {args} {
    return [list start [list 0 [eval SetInput $args]]]
}
$protocol AddMessage start UnsetHandler {args} {
    return [list start [list 0 [eval UnsetHandler $args]]]
}
$protocol AddMessage start InteractiveOutput {args} {
    return [list start [list 0 [eval InteractiveOutput $args]]]
}
# Deprecated -- here only to service old clients
$protocol AddMessage start RemoteRelayOutput {args} {
    return [list start [list 0 [eval RemoteRelayOutput $args]]]
}
$protocol AddMessage start RemoteRelayInput {args} {
    return [list start [list 0 [eval RemoteRelayInput $args]]]
}
$protocol AddMessage start FieldAdjust {args} {
    return [list start [list 0 [eval FieldAdjust $args]]]
}
$protocol AddMessage start ResetThread {args} {
    return [list start [list 0 [eval ResetThread $args]]]
}
$protocol AddMessage start ChangeSolverState {args} {
    return [list start [list 0 [eval ChangeSolverState $args]]]
}
$protocol AddMessage start delete {} {
    global protocol
    return [list start [list 0 [after 100 $protocol Delete]]]
}
Oc_EventHandler New _ $protocol Delete [list after 2000 exit] -oneshot 1
Net_Server New server -protocol $protocol -alias mmSolve2D \
        -log [em_extension ApplicationLog]
$server Start 0

# Delay "nice" of this process until any children are spawned.
Net_Account New a
if {[$a Ready]} {
    Oc_MakeNice
} else {   
    Oc_EventHandler New _ $a Ready Oc_MakeNice -oneshot 1
}

proc SetInput {data host acct pid} {
    Net_Thread New t -hostname $host -accountname $acct -pid $pid \
            -alias {} -log [em_extension ApplicationLog]
    Oc_EventHandler New _ RemoteRelayInput $data \
	    [list RelayInput $data $t] -groups [list $t]
}

proc SetHandler {event data host acct pid cnt} {
    global events
    Net_Thread New t -hostname $host -accountname $acct -pid $pid \
            -alias {} -log [em_extension ApplicationLog]
    Oc_EventHandler DeleteGroup $event-$data-$t
    if {[string match interactive $event]} {
	Oc_EventHandler New _ InteractiveOutput $data \
		[list RelayOutput $data $t] \
		-groups [list $t $event-$data-$t]
    } else {
	set evinfo $events($event)
	set srcName [lindex $evinfo 1]
	upvar #0 $srcName src
	Oc_EventHandler New _ $src $event \
		[list GenericHandler $event $data $cnt $t] \
		-groups [list $t $event-$data-$t]
    }
    global threadrefs
    set threadrefs($event-$data-$t) 1
}

proc UnsetHandler {event data host acct pid} {
    Net_Thread New t -hostname $host -accountname $acct -pid $pid \
            -alias {} -log [em_extension ApplicationLog]
    Oc_EventHandler DeleteGroup $event-$data-$t
    global threadrefs
    catch {unset threadrefs($event-$data-$t)}
    if {![llength [array names threadrefs *-$t]]} {
	$t Delete
    }
}

proc GenericHandler {event data cnt thread} {
    global events
    set evinfo $events($event)
    if {[lindex $evinfo 0]} {
        set srcName [lindex $evinfo 1]
        upvar #0 $srcName src
        set evcount [eval $src [lindex $evinfo 2]]
        if {$evcount % $cnt} {
            return
        }
    }
    RelayOutput $data $thread
}

Oc_EventHandler Bindtags InteractiveOutput InteractiveOutput
proc InteractiveOutput { data } {
    Oc_EventHandler Generate InteractiveOutput $data
}

# Deprecated -- here only to service old clients
proc RemoteRelayOutput { data host acct pid } {
    Net_Thread New t -hostname $host -accountname $acct -pid $pid \
            -alias {} -log [em_extension ApplicationLog]
    RelayOutput $data $t
}

# Send the data indicated by 'data' to the process represented by 'thread'
proc RelayOutput { data thread } {
    # Check to see that 'thread' is ready to send commands to its process
    if {[catch {$thread Ready} ready] || !$ready} {
        # 'thread' isn't ready.  Two possible reasons:
        if {![string match {} [info commands $thread]]} {
            # 1. Thread exists, but just isn't ready yet.
            #    --> Catch its Ready event
            Oc_EventHandler New _ $thread Ready \
                    [list RelayOutput $data $thread] \
                    -oneshot 1 -groups [list $thread]
	}
        # 2. There is no such thread instance --> give up.
    } else {
        # Thread is ready.  Construct and send data.
        global outputs
        set opinfo $outputs($data)
        set srcName [lindex $opinfo 2]
        if {[string length $srcName]} {
            upvar #0 $srcName src
            set datavalue [eval $src [lindex $opinfo 3]]
        } else {
            set datavalue [eval [lindex $opinfo 3]]
        }
        set method [lindex $opinfo 4]
        $thread Send $method $datavalue
    }
}

Oc_EventHandler Bindtags RemoteRelayInput RemoteRelayInput
proc RemoteRelayInput { data } {
    Oc_EventHandler Generate RemoteRelayInput $data
}

# Send request for 'data' to the process represented by 'thread'
proc RelayInput { data thread } {
    # Check to see that 'thread' is ready to send commands to its process
    if {[catch {$thread Ready} ready] || !$ready} {
        # 'thread' isn't ready.  Two possible reasons:
        if {![string match {} [info commands $thread]]} {
            # 1. Thread exists, but just isn't ready yet.
            #    --> Catch its Ready event
            Oc_EventHandler New _ $thread Ready \
                    [list RelayInput $data $thread] \
                    -oneshot 1 -groups [list $thread]
        }
        # 2. There is no such thread instance --> give up.
    } else {
        # Thread is ready.  Construct and send request.
        global inputs 
        set ininfo $inputs($data)
        set qid [$thread Send [lindex $ininfo 1]]    
        Oc_EventHandler New _ $thread Reply$qid \
               [list ReceiveRelayInputReply $thread $data] \
               -oneshot 1 -groups [list $thread]
        # Consider handler for timeout event?
    }
}

proc ReceiveRelayInputReply {thread data} {
    global inputs
    set reply [$thread Get]
    if {![lindex $reply 0]} {
        set ininfo $inputs($data)
        eval [lindex $ininfo 2] [lrange $reply 1 end]
    }
}

# Initialize solver
#$solver Configure -default_control_spec [list -torque 1e-4]
set Bfield_multiplier 1000  ;# Convert T to mT

proc MakeDataTable {} {
    global solver Bfield_multiplier
    set table [list [list Iteration "" [$solver GetODEStepCount]]]
    lappend table [list "Field Updates" {} [$solver GetFieldUpdateCount]]
    lappend table [list "Sim Time" "ns" [expr [$solver GetSimulationTime]*1e9]]
    lappend table [list "Time Step" "ns" [expr [$solver GetTimeStep]*1e9]]
    lappend table [list "Step Size" "" [$solver GetStepSize]]
    set Bfield [$solver GetNomAppliedField]
    lappend table [list Bx mT [expr $Bfield_multiplier * [lindex $Bfield 0]]]
    lappend table [list By mT [expr $Bfield_multiplier * [lindex $Bfield 1]]]
    lappend table [list Bz mT [expr $Bfield_multiplier * [lindex $Bfield 2]]]
    set Bmag [expr sqrt([lindex $Bfield 0]*[lindex $Bfield 0] \
	    + [lindex $Bfield 1]*[lindex $Bfield 1] \
	    + [lindex $Bfield 2]*[lindex $Bfield 2])]
    lappend table [list B mT [expr $Bfield_multiplier * $Bmag]]
    lappend table [list "|m x h|" "" [$solver GetMaxMxH]]
    set mag [$solver GetAveMag]
    lappend table [list Mx/Ms "" [lindex $mag 0]]
    lappend table [list My/Ms "" [lindex $mag 1]]
    lappend table [list Mz/Ms "" [lindex $mag 2]]
    set energies [join [$solver GetEnergyDensities]]
    lappend table [list "Total Energy"      "J/m^3" [lindex $energies 1]]
    lappend table [list "Exchange Energy"   "J/m^3" [lindex $energies 3]]
    lappend table [list "Anisotropy Energy" "J/m^3" [lindex $energies 5]]
    lappend table [list "Demag Energy"      "J/m^3" [lindex $energies 7]]
    lappend table [list "Zeeman Energy"     "J/m^3" [lindex $energies 9]]
    lappend table [list "Max Angle" "deg" [$solver GetMaxAngle]]
    lappend table [list "@Filename:[$solver GetOutputBasename].odt" {} 0]
    return $table
}

proc FieldAdjust { step } {
    global solver 
    if {$step != 0} {
	$solver IncrementFieldStep $step
    }
}

proc ResetThread { {mif {}} } {
    global solver restart_option
    $solver Reset $mif $restart_option
}

proc ChangeSolverState { newstate } {
    global solver
    $solver ChangeSolverState $newstate
}

proc RunEndNotify {} {
    # Send end-of-run message to each thread supporting
    # the DataTable protocol
    set eormsg [list [list "@Filename:" {} 0]]
    foreach thread [Net_Thread Instances] {
	if {[string match DataTable [$thread Protocol]]} {
	    $thread Send DataTable $eormsg
	}
    }
}

vwait forever

